#include "Wire.h"
#include <OneWire.h>

#define MFpin A10
#define relaypin 52
#define button1pin A8
#define button2pin A9

#define MFsamplelength 2048
#define Vibrsamplelength 2048
#define Tempsamplelength 500
#define FFTlength 1024
#define LogFFTlength 10
#define Splength 1024

#define  NUMBER_IS_2_POW_K(x)   ((!((x)&((x)-1)))&&((x)>1))  // x is pow(2, k), k=1,2, ...
#define  FT_DIRECT        -1    // Direct transform.
#define  FT_INVERSE        1    // Inverse transform.

float Re[FFTlength]; //Действительная и мнимая части сигнала для БПФ
float Im[FFTlength];

const int MPU_addr = 0x68; // адрес акселерометра
int16_t accdata[7]; // массив данных
// [accX, accY, accZ, temp, gyrX, gyrY, gyrZ]
// acc - ускорение, gyr - угловая скорость, temp - температура (raw)

OneWire tempSensor(50); // Объект OneWire для датчика температуры на 50 пине
int16_t tempdata[9];

int MF[MFsamplelength], Vibr[Vibrsamplelength], AvMF, AvVibr; 
float SMF[Splength], Svibr[Splength], Svz[Splength];
float Temp[Tempsamplelength];
unsigned long t, ttemp, tb1, tb2;
int i=0, j=0, D=256, itemp=Tempsamplelength;
byte relayState=1, button1State=0, button2State=0, mode=0, outputflag=0, outmode=0, tempflag=0;

void setup() 
{
  // инициализация
  Wire.setClock(400000);
  Wire.begin();
  Wire.beginTransmission(MPU_addr);
  Wire.write(0x6B);  // PWR_MGMT_1 register
  Wire.write(0);     // set to zero (wakes up the MPU-6050)
  Wire.endTransmission(true);

  analogReadResolution(12);
  Serial.begin(9600);

  pinMode(relaypin, OUTPUT);
 }

void loop() 
{
  switch(mode)
  {
    case 0: //ожидание
      if(ttemp < millis()) // измерение температуры
      {
        if(tempflag == 0)
        {
          tempflag=1;
          tempSensor.reset(); // Начинаем взаимодействие со сброса всех предыдущих команд и параметров
          tempSensor.write(0xCC); // Даем датчику DS18b20 команду пропустить поиск по адресу, так как на линии только одно устрйоство 
          tempSensor.write(0x44); // Даем датчику DS18b20 команду измерить температуру. Значение температуры запишется во внутреннюю память датчика
          ttemp = millis() + 1000;
        }
        else
        {
          tempflag=0;
          tempSensor.reset(); // Готовимся получить значение измеренной температуры
          tempSensor.write(0xCC); 
          tempSensor.write(0xBE); // Просим передать значение температуры из регистров
          
          for(i=0;i<2;i++)
          {
            tempdata[i] = tempSensor.read(); // Считываем ответ, младший, затем старший байт
          }
          
          itemp++;
          if(itemp >= Tempsamplelength)
              itemp=0;
          Temp[itemp] =  float((tempdata[1] << 8) | tempdata[0]) * 0.0625; // Формируем итоговое значение
          // 0,0625 - коэффициент, соответсвующий разрешающей способности 12 бит (установленной по умолчанию)
          ttemp = millis() + 4000;
        }  
      }
      
      if(outputflag == 1) // вывод данных в СОМ-порт
      {
        switch(outmode)
        {
          case 0:
          for(i=0;i<500;i++)
            Serial.println((MF[i]-AvMF)*relayState);
          break;
          case 1:
          for(i=0;i<500;i++)
            Serial.println(Vibr[i]-AvVibr);
          break;
          case 2:
          for(i=0;i<500;i++)
            Serial.println(Temp[i]);
          break;
          case 3:
          for(i=0;i<500;i++)
            Serial.println(Temp[itemp]);
          break;
          case 4:
          for(i=0;i<500;i++)
            Serial.println(SMF[i]);
          break;
          case 5:
          for(i=0;i<500;i++)
            Serial.println(Svibr[i]);
          break;
          case 6:
          for(i=0;i<500;i++)
            Serial.println(Svz[i]);
          break;
        }
        outputflag = 0;
      }
    break;

    case 1: //измерение
      relaymode();//функция выбора диапазона
      i = 0;
      while(i < MFsamplelength)
      {
        if(t < micros())
        {
          t = micros() + 1000;
          MF[i] = analogRead(MFpin); // измеряем магнитное поле
          getAccData(3); // получаем данные с акселерометра
          Vibr[i] = accdata[0]*accdata[0]+accdata[1]*accdata[1]+accdata[2]*accdata[2]; // измеряем виброускорение
          i++;
        }
      }
      mode = 2;
      digitalWrite(relaypin, LOW);
    break;

    case 2: //вычисление
      for(i=0;i<Vibrsamplelength;i++) // вычисляем виброускорение
        Vibr[i] = sqrt(Vibr[i]);
      
      //for(i=0;i<MFsamplelength;i++) //тестовый сигнал
        //MF[i] = 1000.0*sin(50.0*(float)i*2.0*3.1415926/1024.0) + 800.0*sin(150.0*(float)i*2.0*3.1415926/1024.0) + 500.0*sin(200.0*(float)i*2.0*3.1415926/1024.0);
      
      AvMF=0; //вычисление средних значений
      AvVibr=0;
      for(i=0;i<MFsamplelength;i++)
        AvMF = AvMF + MF[i];
      for(i=0;i<Vibrsamplelength;i++)
        AvVibr = AvVibr + Vibr[i];
      
      AvMF = AvMF/MFsamplelength;
      AvVibr = AvVibr/Vibrsamplelength;

      for(i=0;i<Splength;i++) //вычисление СПМ методом Уэлча 
      {
        SMF[i] = 0;
        Svibr[i] = 0;
        Svz[i] = 0;
      }
      
      for(j=0;j<MFsamplelength-FFTlength;j+=D) // Вычисление спектров
      {
        for(i=0;i<FFTlength;i++) // Магнитное поле
        {
          Re[i] = (MF[i+j] - AvMF) * float(relayState);
          Im[i] = 0;
        }
        FFT(Re, Im, FFTlength, LogFFTlength, FT_DIRECT); //БПФ магнитного поля
        for(i=0;i<Splength;i++)
          SMF[i] += sqrt(Re[i]*Re[i]+Im[i]*Im[i])/(float)FFTlength; //Вещественная амплитуда спектральных составляющих магнитного поля


        for(i=0;i<FFTlength;i++) // Вибрация
        {
          Re[i] = Vibr[i+j] - AvVibr;
          Im[i] = 0;
        }
        FFT(Re, Im, FFTlength, LogFFTlength, FT_DIRECT); //БПФ вибрации
        for(i=0;i<Splength;i++)
          Svibr[i] += sqrt(Re[i]*Re[i]+Im[i]*Im[i])/(float)FFTlength; //Вещественная амплитуда спектральных составляющих вибрации


        for(i=0;i<FFTlength;i++) // Взаимный спектр
        {
          Re[i] = float(Vibr[i+j]-AvVibr)/1000.0 * float(MF[i+j]-AvMF)*float(relayState); //Ковариация сигналов
          Im[i] = 0;
        }
        FFT(Re, Im, FFTlength, LogFFTlength, FT_DIRECT);
        for(i=0;i<Splength;i++)
          Svz[i] += sqrt(Re[i]*Re[i]+Im[i]*Im[i])/(float)FFTlength;
      }
      
      for(i=0;i<Splength;i++) //Усреднение
      {
        SMF[i] = SMF[i]/((MFsamplelength-FFTlength)/D); 
        Svibr[i] = Svibr[i]/((MFsamplelength-FFTlength)/D); 
        Svz[i] = Svz[i]/((MFsamplelength-FFTlength)/D);
      }
    
    mode = 0;
    outmode = 0;
    outputflag = 1;
    break;
  }
  
    

  //Кнопки
  if(mode == 0)
  {
    switch(button1State)
    {
      case 0: //не нажата
        if(analogRead(button1pin) > 2000)
        {
          button1State = 1; //нажимается
          tb1 = millis();
        }
      break;
      case 1: //нажимается
      if(analogRead(button1pin) > 2000 && millis() - tb1 > 50)
        {
          button1State = 2; //нажата
          //Serial.println("1");
        }
      break;
      case 2: //нажата
        if(analogRead(button1pin) <= 2000)
        {
          button1State = 3; //отпускается
          tb1 = millis();
        }
      break;
      case 3: //отпускается
      if(analogRead(button1pin) <= 2000 && millis() - tb1 > 50)
        {
          button1State = 0; //не нажата
          mode = 1; //Переход в режим измерения
        }
      break;
    }


    switch(button2State)
    {
      case 0: //не нажата
        if(analogRead(button2pin) > 2000)
        {
          button2State = 1; //нажимается
          tb1 = millis();
        }
      break;
      case 1: //нажимается
        if(analogRead(button2pin) > 2000 && millis() - tb1 > 50)
        {
          button2State = 2; //нажата
          //Serial.println("2");
        }
      break;
      case 2: //нажата
        if(analogRead(button2pin) <= 2000)
        {
          button2State = 3; //отпускается
          tb1 = millis();
        }
      break;
      case 3: //отпускается
        if(analogRead(button2pin) <= 2000 && millis() - tb1 > 50)
        {
          button2State = 0; //не нажата
          outputflag = 1;
          outmode++;
          if(outmode > 6)
            outmode = 0;
        }
      break;
    }
  }
}
  


void getAccData(int valuesCount) 
{
  Wire.beginTransmission(MPU_addr);
  Wire.write(0x3B);  // starting with register 0x3B (ACCEL_XOUT_H)
  Wire.endTransmission(false);
  Wire.requestFrom(MPU_addr, valuesCount*2, true); //запрос нужного количества данных (по 2 регистра на величину)
  for (byte i = 0; i < valuesCount; i++) 
  {
    accdata[i] = Wire.read() << 8 | Wire.read();
  }
}

void relaymode()//функция выбора диапазона
{
  int U;
  int kolpred = 0;
  digitalWrite(relaypin, LOW);
  delay(100);
  i = 0;
  while(i< 2000)
  {
    if(t < micros())
    {
      t = micros() + 250;
      U = analogRead(MFpin);
      if(U < 100 || U > 4000)
        kolpred++;
      i++;
    }
  }
  if(kolpred > 10)
  {
    digitalWrite(relaypin, HIGH);
    relayState=10;
  }
  else
  {
    digitalWrite(relaypin, LOW);
    relayState=1; 
  }
  delay(100);
}

bool  FFT(float *Rdat, float *Idat, int N, int LogN, int Ft_Flag) //Функция, вычисляющая БПФ
{
  // parameters error check:
  if((Rdat == NULL) || (Idat == NULL))                  return false;
  if((N > 16384) || (N < 1))                            return false;
  if(!NUMBER_IS_2_POW_K(N))                             return false;
  if((LogN < 2) || (LogN > 14))                         return false;
  if((Ft_Flag != FT_DIRECT) && (Ft_Flag != FT_INVERSE)) return false;

  register int  i, j, n, k, io, ie, in, nn;
  float         ru, iu, rtp, itp, rtq, itq, rw, iw, sr;
  
  static const float Rcoef[14] =
  {  -1.0000000000000000F,  0.0000000000000000F,  0.7071067811865475F,
      0.9238795325112867F,  0.9807852804032304F,  0.9951847266721969F,
      0.9987954562051724F,  0.9996988186962042F,  0.9999247018391445F,
      0.9999811752826011F,  0.9999952938095761F,  0.9999988234517018F,
      0.9999997058628822F,  0.9999999264657178F
  };
  static const float Icoef[14] =
  {   0.0000000000000000F, -1.0000000000000000F, -0.7071067811865474F,
     -0.3826834323650897F, -0.1950903220161282F, -0.0980171403295606F,
     -0.0490676743274180F, -0.0245412285229122F, -0.0122715382857199F,
     -0.0061358846491544F, -0.0030679567629659F, -0.0015339801862847F,
     -0.0007669903187427F, -0.0003834951875714F
  };
  
  nn = N >> 1;
  ie = N;
  for(n=1; n<=LogN; n++)
  {
    rw = Rcoef[LogN - n];
    iw = Icoef[LogN - n];
    if(Ft_Flag == FT_INVERSE) iw = -iw;
    in = ie >> 1;
    ru = 1.0F;
    iu = 0.0F;
    for(j=0; j<in; j++)
    {
      for(i=j; i<N; i+=ie)
      {
        io       = i + in;
        rtp      = Rdat[i]  + Rdat[io];
        itp      = Idat[i]  + Idat[io];
        rtq      = Rdat[i]  - Rdat[io];
        itq      = Idat[i]  - Idat[io];
        Rdat[io] = rtq * ru - itq * iu;
        Idat[io] = itq * ru + rtq * iu;
        Rdat[i]  = rtp;
        Idat[i]  = itp;
      }

      sr = ru;
      ru = ru * rw - iu * iw;
      iu = iu * rw + sr * iw;
    }

    ie >>= 1;
  }

  for(j=i=1; i<N; i++)
  {
    if(i < j)
    {
      io       = i - 1;
      in       = j - 1;
      rtp      = Rdat[in];
      itp      = Idat[in];
      Rdat[in] = Rdat[io];
      Idat[in] = Idat[io];
      Rdat[io] = rtp;
      Idat[io] = itp;
    }

    k = nn;

    while(k < j)
    {
      j   = j - k;
      k >>= 1;
    }

    j = j + k;
  }

  if(Ft_Flag == FT_DIRECT) return true;

  rw = 1.0F / N;

  for(i=0; i<N; i++)
  {
    Rdat[i] *= rw;
    Idat[i] *= rw;
  }

  return true;
}
